{#   Copyright (c) 2015 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#}
# Configuration for {{ loadbalancer_name }}
global
    daemon
    user nobody
    group {{ usergroup }}
    log {{ log_http | default('/dev/log', true)}} local0
    log {{ log_server | default('/dev/log', true)}} local1 notice
    stats socket {{ sock_path }} mode 0666 level user
    {% if loadbalancer.global_connection_limit is defined %}
    maxconn {{ loadbalancer.global_connection_limit }}
    {% endif %}

defaults
    log global
    retries 3
    option redispatch
    timeout connect {{ timeout_connect | default('5000', true) }}
    timeout client {{ timeout_client | default('50000', true) }}
    timeout server {{ timeout_server | default('50000', true) }}

{% block peers %}{% endblock peers %}

{% block proxies %}{% endblock proxies %}
